@@ -8,7 +8,6 @@
 #include <sys/stat.h>
 #ifndef MSDOS
 #include <sys/wait.h>
-#include <sys/param.h>
 #include <unistd.h>
 #endif
 #include <signal.h>
@@ -41,10 +40,15 @@
 #define	X_OK	1		/* Test for execute permission.  */
 #define	F_OK	0		/* Test for existence.  */
 #define L_TREE	1		/* Use different tree style */
+#define DEFARCH 0		/* Default to 8086 code */
 #else
 #define EXESUF
 #endif
 
+#if defined(__minix) || defined(_AIX)
+#define realpath(x,y) 0
+#endif
+
 #define BAS86
 #define BCC86
 
@@ -57,6 +61,7 @@
 #define GCC			"gcc"
 #define LD			"ld86" EXESUF
 #define UNPROTO 		"unproto" EXESUF
+#define OPTIM	 		"copt" EXESUF
 
 #ifdef L_TREE
 #define STANDARD_CRT0_0_PREFIX	"~/lib/"
@@ -66,6 +71,7 @@
 #define DEFAULT_INCLUDE 	"-I~/include"
 #define DEFAULT_LIBDIR0		"-L~/lib/"
 #define DEFAULT_LIBDIR3		"-L~/lib/i386/"
+#define OPTIM_RULES    		"-d~/lib"
 #else
 #define STANDARD_CRT0_0_PREFIX	"~/lib/bcc/i86/"
 #define STANDARD_CRT0_3_PREFIX	"~/lib/bcc/i386/"
@@ -74,6 +80,7 @@
 #define DEFAULT_INCLUDE         "-I~/include"
 #define DEFAULT_LIBDIR0         "-L~/lib/bcc/i86/"
 #define DEFAULT_LIBDIR3         "-L~/lib/bcc/i386/"
+#define OPTIM_RULES    		"-d~/lib/bcc/i86"
 #endif
 
 #ifdef CCC
@@ -123,6 +130,7 @@ PRIVATE struct arg_s asargs = { AS, };
 PRIVATE struct arg_s ccargs = { CC1, CC1_MINUS_O_BROKEN, };
 PRIVATE struct arg_s cppargs = { CPP, };
 PRIVATE struct arg_s unprotoargs = { UNPROTO, TRUE };
+PRIVATE struct arg_s optargs =     { OPTIM };
 #ifdef STANDARD_CRT0_PREFIX
 PRIVATE struct prefix_s crt0_prefix = { STANDARD_CRT0_PREFIX, };
 #endif
@@ -224,6 +232,11 @@ char **argv;
     char *crt0;
 #endif
     char *libc = "-lc";
+#ifdef MSDOS
+    char major_mode = 'd';
+#else
+    char major_mode = 0;
+#endif
     bool_T debug = FALSE;
     bool_T echo = FALSE;
     unsigned errcount = 0;
@@ -240,13 +253,14 @@ char **argv;
     unsigned nifiles = 0;
     unsigned npass_specs;
     bool_T optimize = FALSE;
+    char *optflags = 0;
     char *out_name;
     bool_T profile = FALSE;
     bool_T prep_only = FALSE;
     bool_T prep_line_numbers = FALSE;
     int status;
     char *temp;
-    bool_T patch_exe = FALSE;
+    bool_T patch_exe = FALSE; /* Hackish patch to convert minix i386->OMAGIC */
 
     progname = argv[0];
     addarg(&cppargs, CPPFLAGS);
@@ -258,6 +272,8 @@ char **argv;
     addarg(&ldrargs, "-r");
     addarg(&ldrargs, "-N");	/* GCC uses native objects */
     				/* GCC also uses 386 how to add -3 too ? */
+    addarg(&optargs, "-c!");
+    optflags = stralloc("start");
 #endif
 
 #ifdef L_TREE
@@ -293,7 +309,10 @@ char **argv;
 		prep_line_numbers = FALSE;
 		break;
 	    case 'O':
-		optimize = TRUE;	/* unsupported( arg, "optimize" ); */
+		optimize = TRUE;
+		temp = optflags;
+		optflags=stralloc2(optflags,",86");
+		free(temp);
 		break;
 	    case 'S':
 		cc_only = TRUE;
@@ -396,34 +415,19 @@ char **argv;
 		addarg(&ldargs, arg);
 		break;
 	    case 'M':
-	        switch(arg[2])
+		major_mode=arg[2];
+		break;
+
+	    case 'O':
+		optimize = TRUE;
+		temp=optflags; optflags=stralloc2(optflags,","); free(temp);
+		temp=optflags; optflags=stralloc2(optflags,arg+2); free(temp);
+		if( arg[3] == 0 && ( arg[2] >= '1' && arg[2] <= '3' ))
 		{
-		case 'd': /* DOS compile */
-#ifdef BCC86
-		   bits32 = FALSE;
-#endif
-#ifndef CCC
-		   addarg(&ccargs, "-D__MSDOS__");
-#endif
-		   addarg(&cppargs, "-D__MSDOS__");
-		   addarg(&ldargs, "-d");
-		   addarg(&ldargs, "-s");
-		   addarg(&ldargs, "-T100");
-		   libc= "-ldos";
-		   break;
-		case 'f': /* Caller saves+ax is first arg */
-		   libc= "-lc_f";
-		   addarg(&ccargs, "-f");
-		   addarg(&ccargs, "-c");
-		   break;
-		case 's': /* Standalone executable */
-#ifndef CCC
-		   addarg(&ccargs, "-D__STANDALONE__");
-#endif
-		   addarg(&cppargs, "-D__STANDALONE__");
-		   addarg(&ldargs, "-s");
-		   libc= "-lc_s";
-		   break;
+		   temp=optflags;
+		   optflags=stralloc2(optflags,"86,86");
+		   free(temp);
+		   addarg(&optargs, "-huse16 386");
 		}
 		break;
 	    case 'P':
@@ -476,6 +480,64 @@ char **argv;
     if (errcount != 0)
 	exit(1);
 
+#ifdef BCC86
+    switch(major_mode)
+    {
+    case 'd': /* DOS compile */
+       bits32 = FALSE;
+       libc= "-ldos";
+#ifndef CCC
+       addarg(&ccargs, "-D__MSDOS__");
+#endif
+       addarg(&cppargs, "-D__MSDOS__");
+       addarg(&ldargs, "-d");
+       addarg(&ldargs, "-s");
+       addarg(&ldargs, "-T100");
+       break;
+
+    case 'n': /* Normal Linux-86 */
+       bits32 = FALSE;
+       libc= "-lc";
+       break;
+
+    case 'f': /* Caller saves+ax is first arg */
+       bits32 = FALSE;
+       libc= "-lc_f";
+       addarg(&ccargs, "-f");
+       addarg(&ccargs, "-c");
+       break;
+
+    case 'c': /* Just caller saves, normal C-lib is ok */
+       bits32 = FALSE;
+       libc= "-lc";
+       addarg(&ccargs, "-c");
+       break;
+
+    case 's': /* Standalone executable */
+       bits32 = FALSE;
+       libc= "-lc_s";
+#ifndef CCC
+       addarg(&ccargs, "-D__STANDALONE__");
+#endif
+       addarg(&cppargs, "-D__STANDALONE__");
+       break;
+
+    case 'l': /* Large Linux compile */
+       bits32 = TRUE;
+       libc= "-lc";
+#ifndef CCC
+       addarg(&ccargs, "-D__linux__");
+#endif
+       addarg(&cppargs, "-D__linux__");
+       addarg(&ldargs, "-N"); /* Make OMAGIC */
+       break;
+    case 0:
+       break;
+    default:
+       fatal("Fatal error: illegal -M option given");
+    }
+#endif
+
     if( !aswarn )
        addarg(&asargs, "-w");
     if( patch_exe )
@@ -509,6 +571,14 @@ char **argv;
 #endif
 	    addarg(&ldargs, DEFAULT_LIBDIR0);
     }
+
+    addarg(&optargs, OPTIM_RULES);
+    temp=optflags; optflags=stralloc2(optflags,",end"); free(temp);
+    for(temp=strtok(optflags,","); temp; temp=strtok((char*)0,","))
+    {
+       temp = stralloc2("rules.", temp);
+       addarg(&optargs, temp);
+    }
     addprefix(&exec_prefix, STANDARD_EXEC_PREFIX);
     addprefix(&exec_prefix, STANDARD_EXEC_PREFIX_2);
     cppargs.prog = fixpath(cppargs.prog, &exec_prefix, X_OK);
@@ -519,6 +589,7 @@ char **argv;
     ldrargs.prog = fixpath(ldrargs.prog, &exec_prefix, X_OK);
 #endif
     unprotoargs.prog=fixpath(unprotoargs.prog, &exec_prefix, X_OK);
+    optargs.prog = fixpath(optargs.prog, &exec_prefix, X_OK);
     if (tmpdir == NUL_PTR && (tmpdir = getenv("TMPDIR")) == NUL_PTR)
 #ifdef MSDOS
 	tmpdir = ".";
@@ -612,7 +683,7 @@ char **argv;
 		{
 		    if (prep_only)
 			continue;
-		    if (cc_only)
+		    if (cc_only && !optimize)
 		    {
 			if (f_out != NUL_PTR)
 			    out_name = f_out;
@@ -627,6 +698,25 @@ char **argv;
 		    if (run(in_name, out_name, &ccargs) != 0)
 			continue;
 		    in_name = out_name;
+		    if( optimize )
+		    {
+		        if (cc_only)
+		        {
+			    if (f_out != NUL_PTR)
+			        out_name = f_out;
+			    else
+			    {
+			        out_name = stralloc(basename);
+			        out_name[strlen(out_name) - 1] = 's';
+			    }
+		        }
+		        else
+			    out_name = my_mktemp();
+
+		        if (run(in_name, out_name, &optargs) != 0)
+			    continue;
+		        in_name = out_name;
+		    }
 		    ext = 's';
 		}
 		if (ext == 'S')
@@ -874,7 +964,11 @@ PRIVATE void reset_localprefix()
 
       for(d=s=ptr; d && *s; s=d)
       {
+#ifdef MAXPATHLEN
          char buf[MAXPATHLEN];
+#else
+         char buf[1024];
+#endif
 
 	 free(temp);
          d=strchr(s, ':');
@@ -1129,6 +1223,12 @@ struct arg_s *argp;
     if (verbosity > 4 ) return 0;
 #ifdef MSDOS
     status = spawnv(0, argp->prog, argp->argv+arg0);
+    if( status<0 )
+    {
+	show_who("spawn of ");
+	writes(argp->prog);
+	writesn(" failed");
+    }
 #else
     switch (fork())
     {
